@@ -5,16 +5,17 @@ import asyncio
 import inspect
 import pickle
 import warnings
-from aio_pika import IncomingMessage, Channel, Message, Queue
-from aiormq import PublishError
-from opentelemetry.trace import SpanKind
-from typing import Callable, List, Optional, Set, Dict, NewType, Type, Any, Union, Tuple, get_origin, AsyncIterator
+from typing import Any, AsyncIterator, Callable, Dict, List, NewType, Optional, Set, Tuple, Type, Union, get_origin
 
+from aio_pika import Channel, IncomingMessage, Message, Queue
+from aiormq import PublishError
 from clearcut import context_from_carrier, get_logger_tracer
-from clearcut.otlputils import carrier_from_context
+from clearcut.otlputils import add_event, carrier_from_context
+from opentelemetry.trace import SpanKind
+
 from hatter.amqp import AMQPManager
-from hatter.domain import DecoratedCoroOrGen, RegisteredCoroOrGen, HatterMessage
-from hatter.util import get_substitution_names, create_exchange_queue, flexible_is_subclass, thread_it
+from hatter.domain import GENERATOR_RESPONSE_INDEX_HEADER, DecoratedCoroOrGen, HatterMessage, RegisteredCoroOrGen
+from hatter.util import create_exchange_queue, flexible_is_subclass, get_substitution_names, thread_it
 
 logger, tracer = get_logger_tracer(__name__)
 
@@ -193,6 +194,7 @@ class Hatter:
         concurrency: int = 1,
         autoack: bool = False,
         blocks: bool = False,
+        always_check_rpc_exists: bool = False,
     ) -> Callable[[Union[DecoratedCoroOrGen]], DecoratedCoroOrGen]:
         """
         Registers decorated coroutine (`async def`) or async generator (`async def` with `yield` instead of return) to run when a message is
@@ -220,7 +222,7 @@ class Hatter:
         def decorator(coro_or_gen: DecoratedCoroOrGen) -> DecoratedCoroOrGen:
             # Register this coroutine/async-generator for later listening
             if inspect.iscoroutinefunction(coro_or_gen) or inspect.isasyncgenfunction(coro_or_gen):
-                self._register_listener(coro_or_gen, queue_name, exchange_name, concurrency, autoack, blocks)
+                self._register_listener(coro_or_gen, queue_name, exchange_name, concurrency, autoack, blocks, always_check_rpc_exists)
                 return coro_or_gen
             else:
                 raise ValueError(
@@ -238,6 +240,7 @@ class Hatter:
         concurrency: int,
         autoack: bool,
         blocks: bool,
+        always_check_rpc_exists: bool,
     ):
         """
         Adds function to registry
@@ -250,6 +253,7 @@ class Hatter:
                 concurrency=concurrency,
                 autoack=autoack,
                 blocks=blocks,
+                always_check_rpc_exists=always_check_rpc_exists,
             )
         )
 
@@ -376,7 +380,7 @@ class Hatter:
             logger.debug("Got a message")
             message: IncomingMessage
             try:
-                if message.reply_to is not None and registered_obj.blocks:
+                if registered_obj.always_check_rpc_exists or (message.reply_to is not None and registered_obj.blocks):
                     # Check if response queue exists still. We use "blocks=True" as a limit as those are generally the slower functions.
                     # Because non-existent queues trigger a channel close, we use the REST api to query instead.
                     resp = await self._amqp_manager.rest_client_session.get(f"/api/queues/{self._amqp_manager.vhost}/{message.reply_to}")
@@ -397,6 +401,7 @@ class Hatter:
 
                 # Use context on the message headers
                 with context_from_carrier(message.headers):
+                    add_event("message received")
                     # Build kwargs from fixed...
                     callable_kwargs = dict()
                     callable_kwargs.update(fixed_callable_kwargs)
@@ -448,7 +453,7 @@ class Hatter:
                 logger.info("Cancellation requested during message task instantiation.")
                 raise
             except Exception as e:
-                await self._handle_message_exception(message, e, registered_obj.autoack)
+                await self._handle_message_exception(message, e, not registered_obj.autoack)
 
             logger.debug("Listening again")
 
@@ -557,8 +562,20 @@ class Hatter:
                             "Async iterators which block are not currently supported. Running in main event loop; problems may result."
                         )
 
+                    gen_response_index = 0
                     async for v in registered_obj.coro_or_gen(**callable_kwargs):
-                        await self._handle_return_value(message, v)
+                        await self._handle_return_value(message, v, gen_response_index)
+                        gen_response_index += 1
+
+                    # Send over a StopIteration to tell the listener that we're done
+                    await self.publish(
+                        HatterMessage(
+                            data=StopAsyncIteration(),
+                            destination_queue=message.reply_to,
+                            correlation_id=message.correlation_id,
+                            generator_response_index=gen_response_index,
+                        )
+                    )
 
             # At this point, we're processed the message and sent along new ones successfully. We can ack the original message
             # TODO consider doing all of this transactionally
@@ -571,9 +588,9 @@ class Hatter:
             logger.info("Cancellation requested.")
             raise
         except Exception as e_:
-            await self._handle_message_exception(message, e_, registered_obj.autoack)
+            await self._handle_message_exception(message, e_, not registered_obj.autoack)
 
-    async def _handle_return_value(self, triggering_message: Message, return_val: Any):
+    async def _handle_return_value(self, triggering_message: Message, return_val: Any, generator_response_index: Optional[int] = None):
         if isinstance(return_val, HatterMessage):
             try:
                 logger.debug(f"Publishing {return_val}")
@@ -586,7 +603,10 @@ class Hatter:
             if triggering_message.reply_to is not None:
                 # box it into a HatterMessage
                 rpc_reply = HatterMessage(
-                    data=return_val, destination_queue=triggering_message.reply_to, correlation_id=triggering_message.correlation_id
+                    data=return_val,
+                    destination_queue=triggering_message.reply_to,
+                    correlation_id=triggering_message.correlation_id,
+                    generator_response_index=generator_response_index,
                 )
                 try:
                     await self.publish(rpc_reply)
@@ -602,7 +622,7 @@ class Hatter:
                     f"'reply_to' queue (i.e., RPC pattern)."
                 )
 
-    async def _handle_message_exception(self, message, exception, autoack):
+    async def _handle_message_exception(self, message, exception, nack: bool):
         # Send exception to reply-to queue, if applicable
         # TODO impl more properly. Consider exception type when deciding to requeue. Send to exception handling exchange
         if message.reply_to is not None:
@@ -611,9 +631,9 @@ class Hatter:
             except Exception as e_:
                 logger.exception("Unreplyable exception", exc_info=exception)
                 logger.exception("Error sending exception back to reply-to queue", exc_info=e_)
-        if not autoack:
+        if nack:
             await message.nack(requeue=False)
-        logger.exception("Exception on message")
+        logger.exception("Exception on message", exc_info=exception)
 
     async def _publish_hatter_message(self, msg: HatterMessage, channel: Channel):
         """Intelligently publishes the given message on the given channel"""
@@ -629,12 +649,17 @@ class Hatter:
             # just have to serialize the raw value as needed
             bites: bytes = self.serialize(msg.data)
 
+        headers = carrier_from_context()
+        # Add the gen index if provided
+        if msg.generator_response_index is not None:
+            headers[GENERATOR_RESPONSE_INDEX_HEADER] = msg.generator_response_index
+
         amqp_message = Message(
             body=bites,
             reply_to=msg.reply_to_queue,
             correlation_id=msg.correlation_id,
             priority=msg.priority,
-            headers=carrier_from_context(),
+            headers=headers,
         )
         # TODO other fields like ttl
 
